/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef TRIGMGRCLIENT_HH
#define TRIGMGRCLIENT_HH

#include "TrigWriter.hh"
#include "lmsg/MsgTypes.hh"
#include <string>

namespace trig {
    class TrigBase;
    class Segment;
}

/**  The TrigClient class provides the client API for the Trigger Manager
  *  server. The trigger manager maintains a registry of monitors that 
  *  may produce triggers. The client API requests that it be added to the
  *  registry with the enroll method. The client may then sends one or more 
  *  triggers to the trigger manager with the #sendTrigger()# method. The
  *  trigger data may be routed to an xml file formatted for either the LDAS
  *  database or the online segment database by setting the TRIGGER_FILE or
  *  SEGMENT_FILE environment variables or by specifying the writer mode 
  *  explicitly in the TrigClient constructor, or the #enroll()# method.
  *  @memo Trigger Client API
  *  @author John G. Zweizig
  *  @version 1.5; Modified November 6, 2008
  */
class TrigClient {
public:
    /**  Construct a trigger client.
      *  @memo Construct a trigger client.
      */
    TrigClient(void);

    /**  Construct a trigger client with the specified mode. The trigger client
      *  structure is initialized and if the mode is non-zero, the output 
      *  mode is set and enrollment takes place as described below in the
      *  #enroll()# method.
      *  @memo Construct a trigger client.
      *  @param mode Trigger output mode as specified for the #enroll()# method.
      */
    TrigClient(trig::TrigWriter::trig_mode mode);

    /**  Destroy a trigger client.
      *  @memo Destroy a trigger client.
      */
    virtual ~TrigClient(void);

    /**  Write any remaining table data and close and deletethe writer.
      *  @memo Close a trigger client.
      */
    void close(void);

    /**  The trigger client is initialized for use by this process. There
      *  are presently three trigger output modes available. These are:
      *  <ol>
      *  <li><it>kNone:</it> The writer mode is determined from the run-time 
      *      process environment</li>
      *  <li><it>kMgr:</it>Triggers are sent directly to the trigger manager.
      *      The trigger manager then writes the triggers or segments to an
      *      appropriate database or file system.</li>
      *  <li><it>kWriter:</it>Writes the triggers/segments to an xml file 
      *      formatted appropriately for the LDAS database.</li>
      *  <li><it>kSegWrt:</it>Writes the segments to an xml file 
      *      formatted appropriately for the DQ segment database.</li>
      *  </ol>
      *  If #kNone# mode is specified, the enroll method will attempt to infer
      *  the desired mode from environment variables. If the  #TRIGGER_FILE# 
      *  environment variable is specified, the mode will be set to "kWriter" 
      *  and the output file name will be set to the contents of the variable.
      *  Otherwise, if the #SEGMENT_FILE# variable is set the mode will be 
      *  set to #kSegWrt# and the xml file name will be set to the contents
      *  of the variable name. If neither variable is set, the mode is set to 
      *  kMgr. If nether of these environment variables is set the mode will
      *  be set to #kMgr#. If the mode is set to the #kMgr#, the process will 
      *  register itself with the trigger manager. On successful registration, 
      *  the trigger manager returns a unique process identifier which is used 
      *  to identify the process in subsequent requests.
      *  @memo Enroll process with the Trigger Manager.
      *  @return Error code enumerated in lmsg::Errorlist.hh.
      */
    lmsg::error_type enroll(trig::TrigWriter::trig_mode mode=trig::TrigWriter::kNone);

     /**  flush any existing triggers or segments to the specified output 
       *  file.
       *  @memo get the debug level.
       */
    lmsg::error_type flush(const Time& start, const Time& end);

     /**  Get the debug printout level.
       *  @memo get the debug level.
       */
    lmsg::index_type getDebug(void) const;

    /**  Send a segment to be logged by the trigger manager.
      *  @memo Log a Segment with the trigger manager.
      *  @return Error code enumerated in lmsg::Errorlist.hh.
      *  @param segm Segment to be logged in the metadatabase.
      */
    lmsg::error_type sendSegment(const trig::Segment& segm);

    /**  Send a trigger to the trigger manager.
      *  @memo Log a trigger with the trigger manager.
      *  @return Error code enumerated in lmsg::Errorlist.hh.
      *  @param trigger Trigger to be logged in the metadatabase.
      */
    lmsg::error_type sendTrigger(const trig::TrigBase& trigger);

    /**  Set the debug level.
      *  @memo Set the debug level.
      *  @param level New debugging level.
      */
    void setDebug(lmsg::index_type level);

    /**  Set the table output file name.
      *  @memo Set the table file.
      *  @param name Table file name
      */
    void setTableFile(const char* name);

private:
    /**  This function writes out anydata that is stored in the internal 
      *  database tables.
      */
    void writeTables(void);

private:
    lmsg::index_type  mDebug;
    trig::TrigWriter* mWriter;
    std::string       mTableFile;
};

//======================================  Inline methods.
inline lmsg::index_type 
TrigClient::getDebug(void) const {
    return mDebug;
}

#endif  //   TRIGMGRCLIENT_HH
